import { id } from '../../vendor/izuna/src/index.js'
import { Algebra, Monad } from './interfaces.js'

/** @import { Apply, Chain, Functor, InferredMorphism, Morphism } from './types.js' */

/**
 * @template T
 * @typedef {InferredMorphism<T>} ReaderFn
 */

/** @template T */
export class Reader extends Algebra(Monad) {
  #run

  /** @param {InferredMorphism<T>} run */
  constructor(run) {
    super()

    this.#run = run
  }

  /**
   * @template T
   * @param {T} a
   * @returns {Reader<T>}
   */
  static of(a) {
    return new Reader(/** @type {InferredMorphism<T>} */(_env => a))
  }

  /** @template T */
  static ask() {
    return new Reader(/** @type {InferredMorphism<T>} */(id))
  }

  /**
   * @template U
   * @type {Functor<T>['map']}
   * @param {Morphism<T, U>} f 
   * @returns {Reader<U>}
   */
  map(f) {
    return new Reader(env => f(this.#run(env)))
  }

  /**
   * @type {Chain<T>['chain']}
   * @param {InferredMorphism<T>} f 
   * @returns {Reader<T>}
   */
  chain(f) {
    return new Reader(env => f(this.#run(env)).run(env))
  }

  /**
   * @template U
   * @type {Apply<T>['ap']}
   * @param {Reader<Morphism<T, U>>} other
   * @returns {Reader<U>}
   */
  ap(other) {
    return new Reader(
      env => other.run(env)(this.run(env))
    )
  }

  /**
   * @param {InferredMorphism<T>} f 
   * @returns {Reader<T>}
   */
  local(f) {
    return new Reader(/** @type {InferredMorphism<T>} */(env => this.#run(f(env))))
  }

  /**
   * @template U
   * @param {T} env 
   * @returns {U}
   */
  run(env) {
    return this.#run(env)
  }
}

